@@ -1,14 +1,10 @@
 #define _GNU_SOURCE
-#include <sched.h>
-#include <ctype.h>
 
 #include <haproxy/compat.h>
 #include <haproxy/cpuset.h>
 #include <haproxy/intops.h>
 #include <haproxy/tools.h>
 
-struct cpu_map *cpu_map;
-
 void ha_cpuset_zero(struct hap_cpuset *set)
 {
 #if defined(CPUSET_USE_CPUSET) || defined(CPUSET_USE_FREEBSD_CPUSET)
@@ -21,7 +17,7 @@ void ha_cpuset_zero(struct hap_cpuset *set)
 
 int ha_cpuset_set(struct hap_cpuset *set, int cpu)
 {
-	if (cpu >= ha_cpuset_size())
+	if (cpu < 0 || cpu >= ha_cpuset_size())
 		return 1;
 
 #if defined(CPUSET_USE_CPUSET) || defined(CPUSET_USE_FREEBSD_CPUSET)
@@ -36,7 +32,7 @@ int ha_cpuset_set(struct hap_cpuset *set, int cpu)
 
 int ha_cpuset_clr(struct hap_cpuset *set, int cpu)
 {
-	if (cpu >= ha_cpuset_size())
+	if (cpu < 0 || cpu >= ha_cpuset_size())
 		return 1;
 
 #if defined(CPUSET_USE_CPUSET) || defined(CPUSET_USE_FREEBSD_CPUSET)
@@ -77,14 +73,11 @@ void ha_cpuset_or(struct hap_cpuset *dst, struct hap_cpuset *src)
 
 int ha_cpuset_isset(const struct hap_cpuset *set, int cpu)
 {
-	if (cpu >= ha_cpuset_size())
+	if (cpu < 0 || cpu >= ha_cpuset_size())
 		return 0;
 
 #if defined(CPUSET_USE_CPUSET) || defined(CPUSET_USE_FREEBSD_CPUSET)
-	/* Turn to boolean because musl directly returns the mask as a
-	 * a long instead of an int, hence loses bits 32+.
-	 */
-	return !!CPU_ISSET(cpu, &set->cpuset);
+	return CPU_ISSET(cpu, &set->cpuset);
 
 #elif defined(CPUSET_USE_ULONG)
 	return !!(set->cpuset & (0x1 << cpu));
@@ -141,6 +134,20 @@ void ha_cpuset_assign(struct hap_cpuset *dst, struct hap_cpuset *src)
 #endif
 }
 
+/* returns true if the sets are equal */
+int ha_cpuset_isequal(const struct hap_cpuset *dst, const struct hap_cpuset *src)
+{
+#if defined(CPUSET_USE_CPUSET)
+	return CPU_EQUAL(&dst->cpuset, &src->cpuset);
+
+#elif defined(CPUSET_USE_FREEBSD_CPUSET)
+	return !CPU_CMP(&src->cpuset, &dst->cpuset);
+
+#elif defined(CPUSET_USE_ULONG)
+	return dst->cpuset == src->cpuset;
+#endif
+}
+
 int ha_cpuset_size()
 {
 #if defined(CPUSET_USE_CPUSET) || defined(CPUSET_USE_FREEBSD_CPUSET)
@@ -152,30 +159,6 @@ int ha_cpuset_size()
 #endif
 }
 
-/* Detects CPUs that are bound to the current process. Returns the number of
- * CPUs detected or 0 if the detection failed.
- */
-int ha_cpuset_detect_bound(struct hap_cpuset *set)
-{
-	ha_cpuset_zero(set);
-
-	/* detect bound CPUs depending on the OS's API */
-	if (0
-#if defined(__linux__)
-	    || sched_getaffinity(0, sizeof(set->cpuset), &set->cpuset) != 0
-#elif defined(__FreeBSD__)
-	    || cpuset_getaffinity(CPU_LEVEL_CPUSET, CPU_WHICH_PID, -1, sizeof(set->cpuset), &set->cpuset) != 0
-#else
-	    || 1 // unhandled platform
-#endif
-	    ) {
-		/* detection failed */
-		return 0;
-	}
-
-	return ha_cpuset_count(set);
-}
-
 /* Parse cpu sets. Each CPU set is either a unique number between 0 and
  * ha_cpuset_size() - 1 or a range with two such numbers delimited by a dash
  * ('-'). Each CPU set can be a list of unique numbers or ranges separated by
@@ -230,6 +213,60 @@ int parse_cpu_set(const char **args, struct hap_cpuset *cpu_set, char **err)
 	return 0;
 }
 
+/* Print a cpu-set as compactly as possible and returns the output length.
+ * Returns >size if it cannot emit anything due to length constraints, in which
+ * case it will match what is at least needed to go further, and may return 0
+ * for an empty set. It will emit series of comma-delimited ranges in the form
+ * "beg[-end]".
+ */
+int print_cpu_set(char *output, size_t size, const struct hap_cpuset *cpu_set)
+{
+	struct hap_cpuset set = *cpu_set;
+	int cpus = ha_cpuset_size();
+	int first = -1;
+	int len = 0;
+	int cpu;
+
+	for (cpu = 0; cpu < cpus; cpu++) {
+		if (!ha_cpuset_isset(&set, cpu))
+			continue;
+
+		ha_cpuset_clr(&set, cpu);
+
+		/* check if first of a series*/
+		if (first < 0) {
+			first = cpu;
+			len += snprintf(output + len, size - len, "%d", cpu);
+			if (len >= size)
+				return len + 1;
+
+			/* check if belongs to a range */
+			if (cpu < cpus - 1 && ha_cpuset_isset(&set, cpu + 1)) {
+				if (len + 1 >= size)
+					return len + 2;
+				output[len++] = '-';
+				output[len] = 0;
+			} else
+				first = -1;
+		}
+		else if (cpu >= cpus - 1 || !ha_cpuset_isset(&set, cpu + 1)) {
+			/* end of a series and not first */
+			len += snprintf(output + len, size - len, "%d", cpu);
+			if (len >= size)
+				return len + 1;
+			first = -1;
+		}
+
+		if (first < 0 && ha_cpuset_count(&set) > 0) {
+			if (len + 1 >= size)
+				return len + 2;
+			output[len++] = ',';
+			output[len] = 0;
+		}
+	}
+	return len;
+}
+
 /* Parse a linux cpu map string representing to a numeric cpu mask map
  * The cpu map string is a list of 4-byte hex strings separated by commas, with
  * most-significant byte first, one bit per cpu number.
@@ -261,39 +298,3 @@ void parse_cpumap(char *cpumap_str, struct hap_cpuset *cpu_set)
 		++i;
 	} while (comma);
 }
-
-/* Returns true if at least one cpu-map directive was configured, otherwise
- * false.
- */
-int cpu_map_configured(void)
-{
-	int grp, thr;
-
-	for (grp = 0; grp < MAX_TGROUPS; grp++) {
-		for (thr = 0; thr < MAX_THREADS_PER_GROUP; thr++)
-			if (ha_cpuset_count(&cpu_map[grp].thread[thr]))
-				return 1;
-	}
-	return 0;
-}
-
-/* Allocates everything needed to store CPU information at boot.
- * Returns non-zero on success, zero on failure.
- */
-static int cpuset_alloc(void)
-{
-	/* allocate the structures used to store CPU topology info */
-	cpu_map = calloc(MAX_TGROUPS, sizeof(*cpu_map));
-	if (!cpu_map)
-		return 0;
-
-	return 1;
-}
-
-static void cpuset_deinit(void)
-{
-	ha_free(&cpu_map);
-}
-
-INITCALL0(STG_ALLOC, cpuset_alloc);
-REGISTER_POST_DEINIT(cpuset_deinit);
