@@ -28,6 +28,7 @@
 #include <haproxy/quic_stream.h>
 #include <haproxy/quic_tls.h>
 #include <haproxy/quic_trace.h>
+#include <haproxy/quic_tune.h>
 #include <haproxy/ssl_sock-t.h>
 
 DECLARE_POOL(pool_head_quic_tx_packet, "quic_tx_packet", sizeof(struct quic_tx_packet));
@@ -152,7 +153,7 @@ struct buffer *qc_get_txb(struct quic_conn *qc)
 
 /* Commit a datagram payload written into <buf> of length <length>. <first_pkt>
  * must contains the address of the first packet stored in the payload. When
- * GSO is used, several datagrams can be commited at once. In this case,
+ * GSO is used, several datagrams can be committed at once. In this case,
  * <length> must be the total length of all consecutive datagrams.
  *
  * Caller is responsible that there is enough space in the buffer.
@@ -201,7 +202,7 @@ static int qc_may_build_pkt(struct quic_conn *qc, struct list *frms,
 		 (force_ack || nb_aepkts_since_last_ack >= QUIC_MAX_RX_AEPKTS_SINCE_LAST_ACK));
 
 	TRACE_PRINTF(TRACE_LEVEL_DEVELOPER, QUIC_EV_CONN_PHPKTS, qc, 0, 0, 0,
-	             "%c has_sec=%d cc=%d probe=%d must_ack=%d frms=%d prep_in_flight=%llu cwnd=%llu",
+	             "%c has_sec=%d cc=%d probe=%d must_ack=%d frms=%d prep_in_fligh=%llu cwnd=%llu",
 	             quic_enc_level_char_from_qel(qel, qc),
 	             quic_tls_has_tx_sec(qel), cc, probe, *must_ack, LIST_ISEMPTY(frms),
 	             (ullong)qc->path->prep_in_flight, (ullong)qc->path->cwnd);
@@ -427,8 +428,7 @@ static int qc_send_ppkts(struct buffer *buf, struct ssl_sock_ctx *ctx)
 			}
 			qc->path->in_flight += pkt->in_flight_len;
 			pkt->pktns->tx.in_flight += pkt->in_flight_len;
-			if ((global.tune.options & GTUNE_QUIC_CC_HYSTART) && pkt->pktns == qc->apktns &&
-			    cc->algo->hystart_start_round != NULL)
+			if ((quic_tune.options & QUIC_TUNE_CC_HYSTART) && pkt->pktns == qc->apktns)
 				cc->algo->hystart_start_round(cc, pkt->pn_node.key);
 			if (pkt->in_flight_len)
 				qc_set_timer(qc);
@@ -620,7 +620,6 @@ static int qc_prep_pkts(struct quic_conn *qc, struct buffer *buf,
 			enum quic_pkt_type pkt_type;
 			struct quic_tx_packet *cur_pkt;
 			enum qc_build_pkt_err err;
-			int final_packet = 0;
 
 			TRACE_PROTO("TX prep pkts", QUIC_EV_CONN_PHPKTS, qc, qel);
 
@@ -656,36 +655,32 @@ static int qc_prep_pkts(struct quic_conn *qc, struct buffer *buf,
 			 * datagrams carrying ack-eliciting Initial packets to at least the
 			 * smallest allowed maximum datagram size of 1200 bytes.
 			 */
-
-			/* Padding is activated as soon as Initial data is
-			 * present in the datagram. The only exception is on
-			 * server side if Initial content is not ack-eliciting
-			 * (e.g. a single ACK frame sent as Initial).
-			 */
 			if (qel == qc->iel && (!LIST_ISEMPTY(frms) || probe)) {
 				 /* Ensure that no ack-eliciting packets are sent into too small datagrams */
-				if (end - pos + dglen < QUIC_INITIAL_PACKET_MINLEN) {
+				if (end - pos < QUIC_INITIAL_PACKET_MINLEN) {
 					TRACE_PROTO("No more enough room to build an Initial packet",
 					            QUIC_EV_CONN_PHPKTS, qc);
 					break;
 				}
 
+				/* padding will be set for last QEL, except when probing:
+				 * to build a PING only non coalesced Initial datagram for
+				 * instance when blocked by the anti-amplification limit,
+				 * this datagram MUST be padded.
+				 */
 				padding = 1;
 			}
 
-
-			/* TODO currently it's not possible to emit an ACK and probing data simultaneously (see qc_do_build_pkt()).
-			 * As a side-effect, this could cause coalescing of two packets of the same type which should be avoided.
-			 * To implement this, a new datagram is forced by invokation of qc_txb_store(). This must then be checked
-			 * if padding is required as in this case this will be the last packet of the current datagram.
-			 */
-			if (probe && (must_ack || (qel->pktns->flags & QUIC_FL_PKTNS_ACK_REQUIRED)))
-				final_packet = 1;
-
 			pkt_type = quic_enc_level_pkt_type(qc, qel);
+			/* <paddding> parameter for qc_build_pkt() must not be set to 1 when
+			 * building PING only Initial datagram (a datagram with an Initial
+			 * packet inside containing only a PING frame as ack-eliciting
+			 * frame). This is the case when both <probe> and LIST_EMPTY(<frms>)
+			 * conditions are verified (see qc_do_build_pkt()).
+			 */
 			cur_pkt = qc_build_pkt(&pos, end, qel, tls_ctx, frms,
 			                       qc, ver, dglen, pkt_type, must_ack,
-			                       padding && (!next_qel || final_packet),
+			                       padding && !next_qel && (!probe || !LIST_ISEMPTY(frms)),
 			                       probe, cc, &err);
 			if (!cur_pkt) {
 				switch (err) {
@@ -763,11 +758,11 @@ static int qc_prep_pkts(struct quic_conn *qc, struct buffer *buf,
 			if (probe && qel == qc->ael)
 				break;
 
-			if (LIST_ISEMPTY(frms) && !final_packet) {
+			if (LIST_ISEMPTY(frms)) {
 				/* Everything sent. Continue within the same datagram. */
 				prv_pkt = cur_pkt;
 			}
-			else if (!(global.tune.options & GTUNE_QUIC_NO_UDP_GSO) &&
+			else if (!(quic_tune.options & QUIC_TUNE_NO_UDP_GSO) &&
 			         !(HA_ATOMIC_LOAD(&qc->li->flags) & LI_F_UDP_GSO_NOTSUPP) &&
 			         dglen == qc->path->mtu &&
 			         (char *)end < b_wrap(buf) &&
@@ -779,7 +774,7 @@ static int qc_prep_pkts(struct quic_conn *qc, struct buffer *buf,
 				 */
 				prv_pkt = cur_pkt;
 				dglen = 0;
-				padding = 0;
+
 			}
 			else {
 				/* Finalize current datagram if not all frames sent. */
@@ -972,15 +967,15 @@ int qc_dgrams_retransmit(struct quic_conn *qc)
 
 			qc_prep_hdshk_fast_retrans(qc, &ifrms, &hfrms);
 			TRACE_DEVEL("Avail. ack eliciting frames", QUIC_EV_CONN_FRMLIST, qc, &ifrms);
+			TRACE_DEVEL("Avail. ack eliciting frames", QUIC_EV_CONN_FRMLIST, qc, &hfrms);
 			if (!LIST_ISEMPTY(&ifrms)) {
 				ipktns->tx.pto_probe = 1;
-				qel_register_send(&send_list, qc->iel, &ifrms);
-
-				if (!LIST_ISEMPTY(&hfrms)) {
-					TRACE_DEVEL("Avail. ack eliciting frames", QUIC_EV_CONN_FRMLIST, qc, &hfrms);
+				if (!LIST_ISEMPTY(&hfrms))
 					hpktns->tx.pto_probe = 1;
+
+				qel_register_send(&send_list, qc->iel, &ifrms);
+				if (qc->hel)
 					qel_register_send(&send_list, qc->hel, &hfrms);
-				}
 
 				sret = qc_send(qc, 1, &send_list, 0);
 				qc_free_frm_list(qc, &ifrms);
@@ -1318,70 +1313,6 @@ static inline int quic_write_uint32(unsigned char **buf,
 	return 1;
 }
 
-/* Return the maximum number of bytes we must use to completely fill a
- * buffer with <sz> as size for a data field of bytes prefixed by its QUIC
- * variable-length (may be 0).
- * Also put in <*len_sz> the size of this QUIC variable-length.
- * So after returning from this function we have : <*len_sz> + <ret> <= <sz>
- * (<*len_sz> = { max(i), i + ret <= <sz> }) .
- */
-static inline size_t max_available_room(size_t sz, size_t *len_sz)
-{
-	size_t sz_sz, ret;
-	size_t diff;
-
-	sz_sz = quic_int_getsize(sz);
-	if (sz <= sz_sz)
-		return 0;
-
-	ret = sz - sz_sz;
-	*len_sz = quic_int_getsize(ret);
-	/* Difference between the two sizes. Note that <sz_sz> >= <*len_sz>. */
-	diff = sz_sz - *len_sz;
-	if (unlikely(diff > 0)) {
-		/* Let's try to take into an account remaining bytes.
-		 *
-		 *                  <----------------> <sz_sz>
-		 *  <--------------><-------->  +----> <max_int>
-		 *       <ret>       <len_sz>   |
-		 *  +---------------------------+-----------....
-		 *  <--------------------------------> <sz>
-		 */
-		size_t max_int = quic_max_int(*len_sz);
-
-		if (max_int + *len_sz <= sz)
-			ret = max_int;
-		else
-			ret = sz - diff;
-	}
-
-	return ret;
-}
-
-/* This function computes the maximum data we can put into a buffer with <sz> as
- * size prefixed with a variable-length field "Length" whose value is the
- * remaining data length, already filled of <ilen> bytes which must be taken
- * into an account by "Length" field, and finally followed by the data we want
- * to put in this buffer prefixed again by a variable-length field.
- * <sz> is the size of the buffer to fill.
- * <ilen> the number of bytes already put after the "Length" field.
- * <dlen> the number of bytes we want to at most put in the buffer.
- * Also set <*dlen_sz> to the size of the data variable-length we want to put in
- * the buffer. This is typically this function which must be used to fill as
- * much as possible a QUIC packet made of only one CRYPTO or STREAM frames.
- * Returns this computed size if there is enough room in the buffer, 0 if not.
- */
-static inline size_t max_stream_data_size(size_t sz, size_t ilen, size_t dlen)
-{
-	size_t ret, dlen_sz;
-
-	ret = max_available_room(sz - ilen, &dlen_sz);
-	if (!ret)
-		return 0;
-
-	return ret < dlen ? ret : dlen;
-}
-
 /* Return the length in bytes of <pn> packet number depending on
  * <largest_acked_pn> the largest ackownledged packet number.
  */
@@ -1600,66 +1531,40 @@ static int qc_build_frms(struct list *outlist, struct list *inlist,
 	 * in the switch/case block.
 	 */
 	list_for_each_entry_safe(cf, cfbak, inlist, list) {
-		/* header length, data length, frame length. */
-		size_t hlen, dlen, dlen_sz, avail_room, flen;
+		struct quic_frame *split_frm;
+		size_t flen, split_size;
 
 		if (!room)
 			break;
 
 		switch (cf->type) {
 		case QUIC_FT_CRYPTO:
-			TRACE_DEVEL("          New CRYPTO frame build (room, len)",
-			            QUIC_EV_CONN_BCFRMS, qc, &room, len);
-			/* Compute the length of this CRYPTO frame header */
-			hlen = 1 + quic_int_getsize(cf->crypto.offset_node.key);
-			/* Compute the data length of this CRYPTO frame. */
-			dlen = max_stream_data_size(room, hlen, cf->crypto.len);
-			TRACE_DEVEL(" CRYPTO data length (hlen, crypto.len, dlen)",
-			            QUIC_EV_CONN_BCFRMS, qc, &hlen, &cf->crypto.len, &dlen);
-			if (!dlen)
+			flen = quic_strm_frm_fillbuf(room, cf, &split_size);
+			if (!flen)
 				continue;
 
-			/* CRYPTO frame length. */
-			flen = hlen + quic_int_getsize(dlen) + dlen;
 			TRACE_DEVEL("                 CRYPTO frame length (flen)",
 			            QUIC_EV_CONN_BCFRMS, qc, &flen);
-			/* Add the CRYPTO data length and its encoded length to the packet
-			 * length and the length of this length.
-			 */
-			*len += flen;
-			room -= flen;
-			if (dlen == cf->crypto.len) {
-				/* <cf> CRYPTO data have been consumed. */
-				LIST_DEL_INIT(&cf->list);
-				LIST_APPEND(outlist, &cf->list);
-			}
-			else {
-				struct quic_frame *new_cf;
 
-				new_cf = qc_frm_alloc(QUIC_FT_CRYPTO);
-				if (!new_cf) {
+			if (split_size) {
+				split_frm = quic_strm_frm_split(cf, split_size);
+				if (!split_frm) {
 					TRACE_ERROR("No memory for new crypto frame", QUIC_EV_CONN_BCFRMS, qc);
 					continue;
 				}
-
-				new_cf->crypto.len = dlen;
-				new_cf->crypto.offset_node.key = cf->crypto.offset_node.key;
-				new_cf->crypto.qel = qel;
-				TRACE_DEVEL("split frame", QUIC_EV_CONN_PRSAFRM, qc, new_cf);
-				if (cf->origin) {
+				TRACE_DEVEL("split frame", QUIC_EV_CONN_PRSAFRM, qc, split_frm);
+				if (split_frm->origin)
 					TRACE_DEVEL("duplicated frame", QUIC_EV_CONN_PRSAFRM, qc);
-					/* This <cf> frame was duplicated */
-					LIST_APPEND(&cf->origin->reflist, &new_cf->ref);
-					new_cf->origin = cf->origin;
-					/* Detach the remaining CRYPTO frame from its original frame */
-					LIST_DEL_INIT(&cf->ref);
-					cf->origin = NULL;
-				}
-				LIST_APPEND(outlist, &new_cf->list);
-				/* Consume <dlen> bytes of the current frame. */
-				cf->crypto.len -= dlen;
-				cf->crypto.offset_node.key += dlen;
+				LIST_APPEND(outlist, &split_frm->list);
 			}
+			else {
+				LIST_DEL_INIT(&cf->list);
+				LIST_APPEND(outlist, &cf->list);
+			}
+
+			*len += flen;
+			room -= flen;
+
 			break;
 
 		case QUIC_FT_STREAM_8 ... QUIC_FT_STREAM_F:
@@ -1669,106 +1574,13 @@ static int qc_build_frms(struct list *outlist, struct list *inlist,
 					continue;
 				}
 			}
-			/* Note that these frames are accepted in short packets only without
-			 * "Length" packet field. Here, <*len> is used only to compute the
-			 * sum of the lengths of the already built frames for this packet.
-			 *
-			 * Compute the length of this STREAM frame "header" made a all the field
-			 * excepting the variable ones. Note that +1 is for the type of this frame.
-			 */
-			hlen = 1 + quic_int_getsize(cf->stream.id) +
-				((cf->type & QUIC_STREAM_FRAME_TYPE_OFF_BIT) ? quic_int_getsize(cf->stream.offset) : 0);
-			/* Compute the data length of this STREAM frame. */
-			avail_room = room - hlen;
-			if ((ssize_t)avail_room <= 0)
+
+			flen = quic_strm_frm_fillbuf(room, cf, &split_size);
+			if (!flen)
 				continue;
 
-			TRACE_DEVEL("          New STREAM frame build (room, len)",
-			            QUIC_EV_CONN_BCFRMS, qc, &room, len);
-
-			/* hlen contains STREAM id and offset. Ensure there is
-			 * enough room for length field.
-			 */
-			if (cf->type & QUIC_STREAM_FRAME_TYPE_LEN_BIT) {
-				dlen = QUIC_MIN((uint64_t)max_available_room(avail_room, &dlen_sz),
-				                cf->stream.len);
-				dlen_sz = quic_int_getsize(dlen);
-				flen = hlen + dlen_sz + dlen;
-			}
-			else {
-				dlen = QUIC_MIN((uint64_t)avail_room, cf->stream.len);
-				flen = hlen + dlen;
-			}
-
-			if (cf->stream.len && !dlen) {
-				/* Only a small gap is left on buffer, not
-				 * enough to encode the STREAM data length.
-				 */
-				continue;
-			}
-
-			TRACE_DEVEL(" STREAM data length (hlen, stream.len, dlen)",
-			            QUIC_EV_CONN_BCFRMS, qc, &hlen, &cf->stream.len, &dlen);
 			TRACE_DEVEL("                 STREAM frame length (flen)",
 			            QUIC_EV_CONN_BCFRMS, qc, &flen);
-			/* Add the STREAM data length and its encoded length to the packet
-			 * length and the length of this length.
-			 */
-			*len += flen;
-			room -= flen;
-			if (dlen == cf->stream.len) {
-				/* <cf> STREAM data have been consumed. */
-				LIST_DEL_INIT(&cf->list);
-				LIST_APPEND(outlist, &cf->list);
-
-				qc_stream_desc_send(cf->stream.stream,
-				                    cf->stream.offset,
-				                    cf->stream.len);
-			}
-			else {
-				struct quic_frame *new_cf;
-				struct buffer cf_buf;
-
-				new_cf = qc_frm_alloc(cf->type);
-				if (!new_cf) {
-					TRACE_ERROR("No memory for new STREAM frame", QUIC_EV_CONN_BCFRMS, qc);
-					continue;
-				}
-
-				new_cf->stream.stream = cf->stream.stream;
-				new_cf->stream.buf = cf->stream.buf;
-				new_cf->stream.id = cf->stream.id;
-				new_cf->stream.offset = cf->stream.offset;
-				new_cf->stream.len = dlen;
-				new_cf->type |= QUIC_STREAM_FRAME_TYPE_LEN_BIT;
-				/* FIN bit reset */
-				new_cf->type &= ~QUIC_STREAM_FRAME_TYPE_FIN_BIT;
-				new_cf->stream.data = cf->stream.data;
-				new_cf->stream.dup = cf->stream.dup;
-				TRACE_DEVEL("split frame", QUIC_EV_CONN_PRSAFRM, qc, new_cf);
-				if (cf->origin) {
-					TRACE_DEVEL("duplicated frame", QUIC_EV_CONN_PRSAFRM, qc);
-					/* This <cf> frame was duplicated */
-					LIST_APPEND(&cf->origin->reflist, &new_cf->ref);
-					new_cf->origin = cf->origin;
-					/* Detach this STREAM frame from its origin */
-					LIST_DEL_INIT(&cf->ref);
-					cf->origin = NULL;
-				}
-				LIST_APPEND(outlist, &new_cf->list);
-				cf->type |= QUIC_STREAM_FRAME_TYPE_OFF_BIT;
-				/* Consume <dlen> bytes of the current frame. */
-				cf_buf = b_make(b_orig(cf->stream.buf),
-				                b_size(cf->stream.buf),
-				                (char *)cf->stream.data - b_orig(cf->stream.buf), 0);
-				cf->stream.len -= dlen;
-				cf->stream.offset += dlen;
-				cf->stream.data = (unsigned char *)b_peek(&cf_buf, dlen);
-
-				qc_stream_desc_send(new_cf->stream.stream,
-				                    new_cf->stream.offset,
-				                    new_cf->stream.len);
-			}
 
 			/* TODO the MUX is notified about the frame sending via
 			 * previous qc_stream_desc_send call. However, the
@@ -1778,6 +1590,32 @@ static int qc_build_frms(struct list *outlist, struct list *inlist,
 			 * bufferize and resent the announced data later.
 			 */
 
+			if (split_size) {
+				split_frm = quic_strm_frm_split(cf, split_size);
+				if (!split_frm) {
+					TRACE_ERROR("No memory for new STREAM frame", QUIC_EV_CONN_BCFRMS, qc);
+					continue;
+				}
+
+				TRACE_DEVEL("split frame", QUIC_EV_CONN_PRSAFRM, qc, split_frm);
+				if (split_frm->origin)
+					TRACE_DEVEL("duplicated frame", QUIC_EV_CONN_PRSAFRM, qc);
+				LIST_APPEND(outlist, &split_frm->list);
+				qc_stream_desc_send(split_frm->stream.stream,
+				                    split_frm->stream.offset,
+				                    split_frm->stream.len);
+			}
+			else {
+				LIST_DEL_INIT(&cf->list);
+				LIST_APPEND(outlist, &cf->list);
+				qc_stream_desc_send(cf->stream.stream,
+				                    cf->stream.offset,
+				                    cf->stream.len);
+			}
+
+			*len += flen;
+			room -= flen;
+
 			break;
 
 		default:
@@ -1873,14 +1711,19 @@ static inline uint64_t quic_compute_ack_delay_us(unsigned int time_received,
  * number field in this packet. <pn_len> will also have the packet number
  * length as value.
  *
- * Caller must set <padding> when building the last packet of a datagram which
- * must be at least 1.200 bytes long. Note that padding can also be added
- * automatically to ensure packet size is big enough for header protection
- * sampling.
- *
  * NOTE: This function does not build all the possible combinations of packets
  * depending on its list of parameters. In most cases, <frms> frame list is
  * not empty. So, this function first tries to build this list of frames.
+ * Then some padding is added to this packet if <padding> boolean is set true.
+ * The unique case one wants to do that is when a first Initial packet was
+ * previously built into the same datagram as the currently built one and when
+ * this packet is supposed to pad the datagram, if needed, to build an at
+ * least 1200 bytes long Initial datagram.
+ * If <padding> is not true, if the packet is too short, the packet is also
+ * padded. This is very often the case when no frames are provided by <frms>
+ * and when probing with only a PING frame.
+ * Finally, if <frms> was empty, if <probe> boolean is true this function builds
+ * a PING only packet handling also the cases where it must be padded.
  *
  * Return 1 if succeeded (enough room to buile this packet), O if not.
  */
@@ -1961,7 +1804,6 @@ static int qc_do_build_pkt(unsigned char *pos, const unsigned char *end,
 	/* Build an ACK frame if required. */
 	ack_frm_len = 0;
 	/* Do not ack and probe at the same time. */
-	/* TODO qc_do_build_pkt() must rely on its <probe> argument instead of using QEL <pto_probe> field. */
 	if ((must_ack || (qel->pktns->flags & QUIC_FL_PKTNS_ACK_REQUIRED)) && !qel->pktns->tx.pto_probe) {
 		struct quic_arngs *arngs = &qel->pktns->rx.arngs;
 		BUG_ON(eb_is_empty(&qel->pktns->rx.arngs.root));
@@ -2008,14 +1850,13 @@ static int qc_do_build_pkt(unsigned char *pos, const unsigned char *end,
 				goto comp_pkt_len;
 			}
 
-			/* TODO qc_do_build_pkt() must rely on its <probe> argument instead of using QEL <pto_probe> field. */
 			if (qel->pktns->tx.pto_probe) {
 				/* If a probing packet was asked and could not be built,
 				 * this is not because there was not enough room, but due to
 				 * its frames which were already acknowledeged.
 				 * See qc_stream_frm_is_acked()) called by qc_build_frms().
 				 * Note that qc_stream_frm_is_acked() logs a trace in this
-				 * case mentionning some frames were already acknowledged.
+				 * case mentioning some frames were already acknowledged.
 				 *
 				 * That said, the consequence must be the same: cancelling
 				 * the packet build as if there was not enough room in the
@@ -2057,15 +1898,6 @@ static int qc_do_build_pkt(unsigned char *pos, const unsigned char *end,
 		dglen += len_sz;
 	}
 
-	/* TODO qc_do_build_pkt() must rely on its <probe> argument instead of using QEL <pto_probe> field. */
-	if (qel->pktns->tx.pto_probe && LIST_ISEMPTY(&frm_list) && !cc) {
-		/* Send PING if probing required but no frame avail for sending. */
-		add_ping_frm = 1;
-		len += 1;
-		dglen += 1;
-	}
-
-	/* Handle Initial packet padding if necessary. */
 	if (padding && dglen < QUIC_INITIAL_PACKET_MINLEN) {
 		padding_len = QUIC_INITIAL_PACKET_MINLEN - dglen;
 
@@ -2080,33 +1912,37 @@ static int qc_do_build_pkt(unsigned char *pos, const unsigned char *end,
 			}
 		}
 	}
-
-	/* RFC 9001 5.4.2. Header Protection Sample
-	 *
-	 * To ensure that sufficient data is available for sampling, packets are
-	 * padded so that the combined lengths of the encoded packet number and
-	 * protected payload is at least 4 bytes longer than the sample required
-	 * for header protection. The cipher suites defined in [TLS13] -- other
-	 * than TLS_AES_128_CCM_8_SHA256, for which a header protection scheme
-	 * is not defined in this document -- have 16-byte expansions and 16-
-	 * byte header protection samples. This results in needing at least 3
-	 * bytes of frames in the unprotected payload if the packet number is
-	 * encoded on a single byte, or 2 bytes of frames for a 2-byte packet
-	 * number encoding.
-	 */
-
-	/* Add padding if packet is too small for HP sampling as specified
-	 * above. QUIC TLS algos relies on 16 bytes sample extracted
-	 * QUIC_PACKET_PN_MAXLEN(4) bytes after the PN offset Thus, pn and payload
-	 * must be at least QUIC_PACKET_PN_MAXLEN(4) bytes long, so that the sample
-	 * will be extracted as the AEAD tag.
-	 *
-	 * Note that from here, <len> includes <*pn_len>, the total frame lenghts,
-	 * and QUIC_TLS_TAG_LEN(16).
-	 */
-	if (len < QUIC_PACKET_PN_MAXLEN + QUIC_HP_SAMPLE_LEN) {
-		padding_len = QUIC_PACKET_PN_MAXLEN + QUIC_HP_SAMPLE_LEN - len;
-		len += padding_len;
+	else if (len_frms && len_frms < QUIC_PACKET_PN_MAXLEN) {
+		len += padding_len = QUIC_PACKET_PN_MAXLEN - len_frms;
+	}
+	else if (LIST_ISEMPTY(&frm_list)) {
+		if (qel->pktns->tx.pto_probe) {
+			/* If we cannot send a frame, we send a PING frame. */
+			add_ping_frm = 1;
+			len += 1;
+			dglen += 1;
+			/* Note that only we are in the case where this Initial packet
+			 * is not coalesced to an Handshake packet. We must directly
+			 * pad the datragram.
+			 */
+			if (pkt->type == QUIC_PACKET_TYPE_INITIAL) {
+				if (dglen < QUIC_INITIAL_PACKET_MINLEN) {
+					padding_len = QUIC_INITIAL_PACKET_MINLEN - dglen;
+					padding_len -= quic_int_getsize(len + padding_len) - len_sz;
+					len += padding_len;
+				}
+			}
+			else {
+				/* Note that +1 is for the PING frame */
+				if (*pn_len + 1 < QUIC_PACKET_PN_MAXLEN)
+					len += padding_len = QUIC_PACKET_PN_MAXLEN - *pn_len - 1;
+			}
+		}
+		else {
+			/* If there is no frame at all to follow, add at least a PADDING frame. */
+			if (!ack_frm_len && !cc)
+				len += padding_len = QUIC_PACKET_PN_MAXLEN - *pn_len;
+		}
 	}
 
 	if (pkt->type != QUIC_PACKET_TYPE_SHORT && !quic_enc_int(&pos, end, len))
@@ -2179,7 +2015,7 @@ static int qc_do_build_pkt(unsigned char *pos, const unsigned char *end,
 		goto no_room;
 	}
 
-	BUG_ON(qel->pktns->tx.pto_probe && !cc &&
+	BUG_ON(qel->pktns->tx.pto_probe &&
 	       !(pkt->flags & QUIC_FL_TX_PACKET_ACK_ELICITING));
 	/* If this packet is ack-eliciting and we are probing let's
 	 * decrement the PTO probe counter.
