@@ -320,7 +320,7 @@ ssize_t ring_write(struct ring *ring, size_t maxlen, const struct ist pfx[], siz
 
 	vp_ring_to_data(&v1, &v2, ring_area, ring_size, head_ofs, tail_ofs);
 
-	while (vp_size(v1, v2) + needed + 1 + 1 > ring_size) {
+	while (vp_size(v1, v2) > ring_size - needed - 1 - 1) {
 		/* we need to delete the oldest message (from the end),
 		 * and we have to stop if there's a reader stuck there.
 		 * Unless there's corruption in the buffer it's guaranteed
@@ -339,7 +339,7 @@ ssize_t ring_write(struct ring *ring, size_t maxlen, const struct ist pfx[], siz
 
 	/* now let's update the buffer with the new tail if our message will fit */
 	new_tail_ofs = tail_ofs;
-	if (vp_size(v1, v2) + needed + 1 + 1 <= ring_size) {
+	if (vp_size(v1, v2) <= ring_size - needed - 1 - 1) {
 		vp_data_to_ring(v1, v2, ring_area, ring_size, &head_ofs, &tail_ofs);
 
 		/* update the new space in the buffer */
@@ -526,10 +526,10 @@ int ring_attach_cli(struct ring *ring, struct appctx *appctx, uint flags)
 		               "Sorry, too many watchers (" TOSTR(RING_MAX_READERS) ") on this ring buffer. "
 		               "What could it have so interesting to attract so many watchers ?");
 
-	if (!appctx->io_handler)
-		appctx->io_handler = cli_io_handler_show_ring;
-	if (!appctx->io_release)
-                appctx->io_release = cli_io_release_show_ring;
+	if (!appctx->cli_ctx.io_handler)
+		appctx->cli_ctx.io_handler = cli_io_handler_show_ring;
+	if (!appctx->cli_ctx.io_release)
+                appctx->cli_ctx.io_release = cli_io_release_show_ring;
 
 	memset(ctx, 0, sizeof(*ctx));
 	ctx->ring  = ring;
@@ -550,7 +550,8 @@ int ring_attach_cli(struct ring *ring, struct appctx *appctx, uint flags)
  * processed by the function (even when the function returns 0)
  */
 int ring_dispatch_messages(struct ring *ring, void *ctx, size_t *ofs_ptr, size_t *last_ofs_ptr, uint flags,
-                           ssize_t (*msg_handler)(void *ctx, struct ist v1, struct ist v2, size_t ofs, size_t len),
+                           ssize_t (*msg_handler)(void *ctx, struct ist v1, struct ist v2, size_t ofs, size_t len, char delim),
+                           char delim,
                            size_t *processed)
 {
 	size_t head_ofs, tail_ofs, prev_ofs;
@@ -644,7 +645,7 @@ int ring_dispatch_messages(struct ring *ring, void *ctx, size_t *ofs_ptr, size_t
 
 		BUG_ON(msg_len + cnt + 1 > vp_size(v1, v2));
 
-		copied = msg_handler(ctx, v1, v2, cnt, msg_len);
+		copied = msg_handler(ctx, v1, v2, cnt, msg_len, delim);
 		if (copied == -2) {
 			/* too large a message to ever fit, let's skip it */
 			goto skip;
@@ -695,7 +696,6 @@ int ring_dispatch_messages(struct ring *ring, void *ctx, size_t *ofs_ptr, size_t
 int cli_io_handler_show_ring(struct appctx *appctx)
 {
 	struct show_ring_ctx *ctx = appctx->svcctx;
-	struct stconn *sc = appctx_sc(appctx);
 	struct ring *ring = ctx->ring;
 	size_t last_ofs;
 	size_t ofs;
@@ -703,13 +703,14 @@ int cli_io_handler_show_ring(struct appctx *appctx)
 
 	MT_LIST_DELETE(&appctx->wait_entry);
 
-	ret = ring_dispatch_messages(ring, appctx, &ctx->ofs, &last_ofs, ctx->flags, applet_append_line, NULL);
+	ret = ring_dispatch_messages(ring, appctx, &ctx->ofs, &last_ofs, ctx->flags, applet_append_line,
+				     (ctx->flags & RING_WF_END_ZERO) ? 0 : '\n', NULL);
 
 	if (ret && (ctx->flags & RING_WF_WAIT_MODE)) {
 		/* we've drained everything and are configured to wait for more
 		 * data or an event (keypress, close)
 		 */
-		if (!sc_oc(sc)->output && !(sc->flags & SC_FL_SHUT_DONE)) {
+		if (!b_data(&appctx->inbuf) && !se_fl_test(appctx->sedesc, SE_FL_SHW)) {
 			/* let's be woken up once new data arrive */
 			MT_LIST_APPEND(&ring->waiters, &appctx->wait_entry);
 			ofs = ring_tail(ring);
@@ -724,9 +725,11 @@ int cli_io_handler_show_ring(struct appctx *appctx)
 			ret = 0;
 		}
 		/* always drain all the request */
-		co_skip(sc_oc(sc), sc_oc(sc)->output);
+		b_reset(&appctx->inbuf);
+		applet_fl_clr(appctx, APPCTX_FL_INBLK_FULL);
 	}
 
+	applet_will_consume(appctx);
 	applet_expect_no_data(appctx);
 	return ret;
 }
